#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 18 13:43:32 2018

@author: lmasson
"""
import os
import numpy as np
import itertools
import simplif_impact

class Result_empty:
    def __init__(self, id, value):
        self.id=id
        self.val=value

class Result_interv:
    def __init__(self, id, name_interv, value):
        self.id=id
        self.name_interv=name_interv
        self.val=value
        
class Result_comb:
   def __init__(self, id, names_intervs, value)        :
      self.id=id
      self.names_intervs=names_intervs
      self.val=value

class Result_impact_global:
   def __init__(self, id, name):
      self.id=id
      self.name=name

def check_impact_main(model_name, states_structure, interventions_structure, fct):
   results_empty_structure = []
   results_interv_structure = []
   results_combinations_structure = []
   check_impact_empty_state(states_structure, interventions_structure, results_empty_structure, fct)
   #For states that are OK when empty, check impact of interventions
   for s in states_structure:
      for r in results_empty_structure:
         if s.id == r.id:
            if r.val=="true": 
               nb_interv_in_state=np.sum(s.state_interv)
               if nb_interv_in_state > 0:
                  for k in range(len(s.state_interv)):
                     if s.state_interv[k]==1:
                        for i in interventions_structure:
                           if i.id == k:
                              check_impact_with_intervs(s, i, results_interv_structure, fct)
   #For states in which every interv alone is OK: check combinations of intervs                                 
   for s in states_structure:
      s_already_handled=0
      for r in results_interv_structure:
         if r.id == s.id:
            if r.val == "false":
               s_already_handled=1
               break
      if s_already_handled != 1:
        check_impact_combinations_of_intervs(s, interventions_structure, results_combinations_structure, fct)
   if any(r.val=='false' for r in results_empty_structure) or any(r.val=='false' for r in results_interv_structure) or any(r.val=='false' for r in results_combinations_structure):
      #Print global result
      print_results(results_empty_structure, results_interv_structure, results_combinations_structure, fct)
      ##Simplify result
      simplif_impact.simplif_impact_main(model_name, states_structure, results_empty_structure, results_interv_structure, results_combinations_structure, fct)
   if all(r.val=='true' for r in results_empty_structure) and all(r.val=='true' for r in results_interv_structure) and all(r.val=='true' for r in results_combinations_structure):
      print("No problem found")
   store_permissiveness_impact_in_file(results_empty_structure, results_interv_structure, results_combinations_structure, model_name, fct, interventions_structure)
#########################################################################
def check_impact_empty_state(states_structure, interventions_structure, results_empty_structure, fct):
   #Re-define the strategy with no interv in the considered state
   for s in states_structure:
      script=open("script_ctlcheck_empty_state","w")
      state_empty=open("state_empty","w")
      model_copy=open("model_copy.smv","r")
      model_and_part_strategy=open("model_and_part_strategy.smv", "w")
      model_and_part_strategy.write(model_copy.read())
      #Write the new strategy (no interv in state s)
      for i in interventions_structure:
         write_strategy_empty_in_s(i, s, state_empty)
      #Write the ctl property to check
      state_empty.write("CTLSPEC NAME empty_" +str(s.id)+ " := EF(flag_st_" +str(s.id)+ ") -> EF(flag_st_" +str(s.id)+ " & EX(EF("+fct.name+"))) ;")
      state_empty.close()        
      #Write the nusmv script
      script.write("go \n")
      script.write("check_ctlspec -P \"empty_"+str(s.id)+"\"\n")
      script.write("quit\n")
      script.close()
      state_empty=open("state_empty","r")
      model_and_part_strategy.write(state_empty.read())
      state_empty.close()
      model_and_part_strategy.close()
      #!!Change this part for the use of the library pynusmv!!
      os.system("NuSMV -source script_ctlcheck_empty_state model_and_part_strategy.smv > result")
      #!!
      store_results_empty(s, results_empty_structure)
      open("state_empty","w").close()
      open("model_and_part_strategy.smv", "w").close()
      open("result", "w").close
    #print("We checked every state empty")
    
#########################################################################
def write_strategy_empty_in_s(i, s, state_empty):     
    state_empty.write("DEFINE flag_"+i.name+" := ")
#          print("DEFINE flag_"+i.name+" := ")
    first=1
    empty=1
    for i_st in range(len(i.flag)):
        if i.flag[i_st]==1 and i_st!=int(s.id):
            if first==1:
                state_empty.write("flag_st_"+str(i_st))
#                      print("flag_st_"+str(i_st))
                first=0
            else:
                state_empty.write(" | flag_st_"+str(i_st))
#                      print(" | flag_st_"+str(i_st))
            empty=0
    if empty==1:
        state_empty.write("FALSE")
#              print("FALSE")
    state_empty.write(" ;\n")
#          print(" ;\n")


#########################################################################
def store_results_empty(s, results_empty_structure):
   result=open("result", "r")
   lines=result.readlines()
   result.close()
   line=lines[-1]
   if line.split(" is ")[-1].rstrip()=="true":
       result_empty=Result_empty(s.id, "true")
       results_empty_structure.append(result_empty)
   else:
       result_empty=Result_empty(s.id, "false")
       results_empty_structure.append(result_empty)

#########################################################################        
def check_impact_with_intervs(s, i, results_interv_structure, fct):
   script=open("script_ctlcheck_state_interv","w")
   script.write("go\n")
   script.write("check_ctlspec -p \"EF(flag_st_"+s.id+" & "+i.name+".activ) -> EF(flag_st_"+s.id+" & "+i.name+".activ & EX(EF("+fct.name+")))\"\n")
   script.write("quit")
   script.close()                                   
   #!!Change this part for the use of the library pynusmv!!                                   
   os.system("NuSMV -source script_ctlcheck_state_interv model_and_strategy.smv > result")
   store_results_interv(s, i, results_interv_structure)
   #!!                                      
   open("result", "w").close                                   

#########################################################################
def store_results_interv(s, i, results_interv_structure):
     result=open("result", "r")
     lines=result.readlines()
     result.close()
     line=lines[-1]
     if line.split(" is ")[-1].rstrip()=="true":
         result_interv=Result_interv(s.id, i.name, "true")
         results_interv_structure.append(result_interv)
     else:
         result_interv=Result_interv(s.id, i.name, "false")
         results_interv_structure.append(result_interv)

#########################################################################
def check_impact_combinations_of_intervs(s, interventions_structure, results_combinations_structure, fct): 
   combinations_s=combinations(s, interventions_structure)
   for k in combinations_s:
#      print("Check of combinations:"+k)
      script=open("script_ctlcheck_combinations", "w")
      script.write("go\ncheck_ctlspec -p \"EF(flag_st_"+s.id+" & "+k+") -> EF(flag_st_"+s.id+" & "+k+" & EX(EF("+fct.name+")))\"\nquit\n")
      script.close()
      os.system("NuSMV -source script_ctlcheck_combinations model_and_strategy.smv > result")
      store_results_combinations(s, k, results_combinations_structure)
      open("result", "w").close()


#########################################################################
# Creates an array with all the possible combinations of interventions (i1 & i2, i1 & i3, etc...)
def combinations(s, interventions_structure):
   combinations = []
   names_intervs_in_s = create_list_names_intervs_activ_in_state(s, interventions_structure)
   for j in range(2,len(names_intervs_in_s)+1):
      for k in itertools.combinations(names_intervs_in_s, j):
         combinations.append(' & '.join(k))
   return combinations
   
#########################################################################
#Just adds ".activ" to the interventions names   
def create_list_names_intervs_activ_in_state(s, interventions_structure):
   names_intervs_in_s=[]
   for k in range(len(s.state_interv)):
      if s.state_interv[k]==1:
         for i in interventions_structure:
            if i.id==k:
               name=i.name+".activ"
               names_intervs_in_s.append(name)
#   print(list(names_intervs_in_s))
   return names_intervs_in_s

#########################################################################
def store_results_combinations(s, combinations_s_index, results_combinations_structure):
   result=open("result", "r")
   lines=result.readlines()
   result.close()
   line=lines[-1]
   if line.split(" is ")[-1].rstrip()=="true":
      result_combination=Result_comb(s.id, combinations_s_index, "true")
      results_combinations_structure.append(result_combination)
   else:
      result_combination=Result_comb(s.id, combinations_s_index, "false")
      results_combinations_structure.append(result_combination)
   return results_combinations_structure

#########################################################################                                   
def print_results(results_empty_structure, results_interv_structure, results_combinations_structure, fct):
   for r in results_empty_structure:
      if r.val=="false":
         print(fct.name+" not reachable from state "+r.id+ " without intervention")
#      else: print(fct.name+" reachable from state "+r.id+ " without intervention")
   for r in results_interv_structure:
      if r.val=="false":
         print(fct.name+" not reachable from state "+r.id+" with interv "+r.name_interv)
#      else: print(fct.name+" reachable from state "+r.id+" with interv "+r.name_interv)
   for r in results_combinations_structure:
      if r.val == "false":
         print(fct.name+" not reachable from state "+ r.id+" with "+r.names_intervs)
#      else: print(fct.name+" reachable from state "+r.id+" with "+r.names_intervs)
         
#########################################################################
def store_permissiveness_impact_in_file(results_empty_structure, results_interv_structure, results_combinations_structure, model_name, fct, interventions_structure):
   model_pi=open(model_name+".pi", "a")
   model_pi.write("1\n")
   model_pi.write(fct.name+"\n")
   for r in results_empty_structure:
      if r.val=="false":
         model_pi.write(r.id+" 0\n")
   for r in results_interv_structure:
      if r.val=="false":
         for i in interventions_structure:
            if i.name==r.name_interv:
               model_pi.write(str(r.id)+" "+str(i.id)+"\n")         
   for r in results_combinations_structure:
      if r.val == "false":
         model_pi.write(r.id+" ")
         first=1
         for i in interventions_structure:
            if i.name in r.names_intervs:         
               if first==1:
                  first=0
                  model_pi.write(str(i.id))  
               if first!=1:
                  model_pi.write(","+str(i.id))  
   model_pi.close()