## Helper functions
##Deletes the lines of comment from the file
def get_relevant_lines( lines ):
    relevant = []
    for l in lines:
        if l.startswith("DEFINE flag_"):
            relevant.append(l) ##add l to list "relevant"
    return relevant

def is_state_definition( line ):
    return line.startswith("DEFINE flag_st_")

def assign_lefthand_side( line ):
    return line.split(" := ")[0]

def assign_righthand_side( line ):
    rhs = line.split(" := ")[-1].rstrip()##Removes the "end of line" character
    rhs_sans_semicolon = rhs.split(";")[0]
    return rhs_sans_semicolon
        
def state_id( line ):
    lhs = assign_lefthand_side(line)
    id = lhs.split("_")[-1]
    return id

def state_variables( line ):
    rhs = assign_righthand_side(line)
    variables = rhs.split("&")
    variable_structures = [Variable(v) for v in variables]
    return variable_structures

def intervention_name( line ):
    lhs = assign_lefthand_side(line)
    return lhs[len("DEFINE flag_"):]
    
def states_using_intervention( line ):
    rhs = assign_righthand_side(line)
    state_flags = rhs.split(" | ")
    states = []
    for f in state_flags:
       if f!="FALSE ":
        state = int(f.split("_")[-1])
        states.append(state)
    return states

def get_number_of_warning_states( lines ):
    for l in lines:
        if l.startswith("--Number of warning states:"):
            number = int(l.split(':')[-1])
            return number
    # We expect to have some warning states
    assert False, "Could not find any warning states"

def get_number_of_interventions( lines ):
    for l in lines:
        if l.startswith("--Number of interventions:"):
            number = int(l.split(':')[-1])
            return number
    # We expect to have some warning states
    assert False, "Could not find any intervention"


## Data Structures

class State:
    def __init__(self, id, variables, state_interventions):
        self.id = id
        self.variables = variables
        self.state_interv = state_interventions

class Variable:
    def __init__(self, line):
        lhs = line.split("=")[0]
        rhs = line.split("=")[-1]
        self.name = lhs
        self.value = rhs

class Intervention:
    def __init__(self, name, id, intervention_states, warning_states_num):
        flag = [0]*warning_states_num # initialize array of 0's
        for s in intervention_states:
            flag[s] = 1
        self.name = name
        self.id = id
        self.flag = flag

def parse_strategy(lines):
## Execute Parse
    warning_states_num = get_number_of_warning_states(lines)
    interv_num = get_number_of_interventions(lines)
    relevant_lines=get_relevant_lines(lines)

    state_lines = []
    intervention_lines = []
    for l in relevant_lines:
        if is_state_definition(l):
            state_lines.append(l)
        else:
            intervention_lines.append(l)
    
    interventions_structure = []
    id_interv=0
    for l in intervention_lines:
        name = intervention_name(l)
        id = id_interv
        id_interv += 1
        i_states = states_using_intervention(l) ##list of the numbers of the states using the intervention
        intervention = Intervention(name, id, i_states, warning_states_num)
        interventions_structure.append(intervention)
    
    states_structure = []
    for l in state_lines:
        id = state_id(l)
        vars = state_variables(l)
        state_interv = [0]*interv_num
        for i in interventions_structure:
            if i.flag[int(id)]==1:
                state_interv[i.id]=1
        state = State(id, vars, state_interv)
        states_structure.append(state)   
        
    return relevant_lines, state_lines, states_structure, interventions_structure
    
    

## Print Data Structures
def print_strategy_data(states_structure, interventions_structure):
   for s in states_structure:
      print("Warning State "+s.id)
      for v in s.variables:
         print("  variable: ")
         print("    name: "+v.name)
         print("    value: "+v.value)
      print("Interventions in state " + s.id + " are " + str(s.state_interv))
      print("")
    
   for i in interventions_structure:
      print("Intervention "+i.name)
      print("  warning state flag: "+str(i.flag))
   print("")
