#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 27 17:36:01 2018

@author: lmasson
"""

import parse_rsm
from subprocess import call
import utils
import copy
import os

path=os.environ.copy()

class FormulaLine:
   def __init__(self, exists_or, exists_and, names, values):
      self.exists_or=exists_or
      self.exists_and=exists_and
      self.names = names
      self.values = values

def simplif_impact_main(model_name, states_structure, results_empty_structure, results_interv_structure, results_combinations_structure, fct):
   (variables_structure, nb_var, names_intervs, nb_interv, nb_states)=parse_rsm.parse_rsm(model_name)
   simplif_impact_declare_variables(variables_structure, names_intervs)
   simplif_impact_declare_impact_formula(states_structure, results_empty_structure, results_interv_structure, results_combinations_structure)
   impact_simple=simplif_impact_simplify(nb_var, nb_interv)
   formula_lines_structure = simplif_impact_parse_result_smt(impact_simple)
   simplif_strategy_write_result_smv(formula_lines_structure, fct.name, model_name, names_intervs)
#   parse_result=utils.parse_smt('impact_to_implify.smt')
#   terms_smv=utils.write_smv_from_smt(parse_result)

##Write variables declaration in file vars_declaration.smt
def simplif_impact_declare_variables(variables_structure, names_intervs):
   vars_declaration=open("vars_declaration.smt","w")
   for v in variables_structure:
      vars_declaration.write("(declare-datatypes () (( "+v.name+"TYPE ")
      for n in range(0,int(v.bound)+1):
         vars_declaration.write(v.name+str(n)+" ")
      vars_declaration.write(")))\n")
      vars_declaration.write("(declare-fun "+v.name+" () "+v.name+"TYPE )\n")
   vars_declaration.close()
   interv_declaration=open("interv_declaration.smt", "w")
   for n in range(len(names_intervs)):
      interv_declaration.write("(declare-datatypes () (( "+names_intervs[n].rstrip()+"TYPE "+names_intervs[n].rstrip()+"0 "+names_intervs[n].rstrip()+"1 )))\n") #interv value = 0 if not activ
      interv_declaration.write("(declare-fun "+names_intervs[n].rstrip()+" () "+names_intervs[n].rstrip()+"TYPE )\n")
   interv_declaration.close()

def simplif_impact_declare_impact_formula(states_structure, results_empty_structure, results_interv_structure, results_combinations_structure):
   open("impact_declaration.smt","w").close()
   impact_declaration=open("impact_declaration.smt", "w")
   impact_declaration.write("(assert \n")
   impact_empty_smt=[]
   impact_empty_smt_lines=[]
   impact_interv_smt=[]
   impact_interv_smt_lines=[]
   impact_combinations_smt=[]
   impact_combinations_smt_lines=[]
   for r in results_empty_structure:
      impact_empty_smt=[]
      if r.val=="false":
         for s in states_structure:
            if s.id==r.id:
               for v  in  s.variables:
                  impact_empty_smt.append("(= "+v.name.split(".v")[0]+" "+v.name.split(".v")[0]+v.value+") ")
         if len(impact_empty_smt)>1:
            impact_empty_smt_lines.append("\t\t(and "+(" ").join(impact_empty_smt)+")\n")
         if len(impact_empty_smt)==1:
            impact_empty_smt_lines.append(impact_empty_smt[0]+"\n")
#   print(list(impact_empty_smt_lines))
   for r in results_interv_structure:
      impact_interv_smt=[]
      if r.val=="false":
         for s in states_structure:
            if s.id==r.id:
               for v  in  s.variables:
                  impact_interv_smt.append("(= "+v.name.split(".v")[0]+" "+v.name.split(".v")[0]+v.value+") ")
         impact_interv_smt.append("(= "+r.name_interv+" "+r.name_interv+"1 )")
         if len(impact_interv_smt)>1:
            impact_interv_smt_lines.append("\t\t(and "+(" ").join(impact_interv_smt)+")\n")
         if len(impact_interv_smt)==1:
            impact_interv_smt_lines.append(impact_interv_smt[0]+"\n"  ) 
#   print(list(impact_interv_smt_lines))
   for r in results_combinations_structure:
      impact_combinations_smt=[]
      if r.val=="false":
         for s in states_structure:
            if s.id==r.id:
               for v  in  s.variables:
                  impact_combinations_smt.append("(= "+v.name.split(".v")[0]+" "+v.name.split(".v")[0]+v.value+") ")
         names_intervs_list=r.names_intervs.split(" & ")
         for j in range(len(names_intervs_list)):
#            print(names_intervs_list[j])
            impact_combinations_smt.append("(= "+names_intervs_list[j].split(".activ")[0]+" "+names_intervs_list[j].split(".activ")[0]+"1 )")
         if len(impact_combinations_smt)>1:
            impact_combinations_smt_lines.append("\t\t(and "+(" ").join(impact_combinations_smt)+")\n")
         if len(impact_combinations_smt)==1:
            impact_combinations_smt_lines.append(impact_combinations_smt[0] +"\n")
#   print(list(impact_combinations_smt_lines))
   if (len(impact_empty_smt)+len(impact_interv_smt)+len(impact_combinations_smt))>1:
      impact_declaration.write("\n\t(or\n")
   if len(impact_empty_smt_lines)!=0:
      for l in impact_empty_smt_lines:
         impact_declaration.write(l)
   if len(impact_interv_smt_lines)!=0:
      for l in impact_interv_smt_lines:
         impact_declaration.write(l)
   if len(impact_combinations_smt_lines)!=0:
      for l in impact_combinations_smt_lines:
         impact_declaration.write(l)
   if (len(impact_empty_smt)+len(impact_interv_smt)+len(impact_combinations_smt))>1:
      impact_declaration.write("\t)\n")
   impact_declaration.write(")\n")
   impact_declaration.write("(apply ctx-solver-simplify)\n")
   impact_declaration.close()
   
def simplif_impact_simplify(nb_var, nb_interv):
#   print("----------------------Begin simplification-------------------------")
   vars_declaration=open("vars_declaration.smt","r")
   interv_declaration=open("interv_declaration.smt", "r")
   impact_declaration=open("impact_declaration.smt", "r")
   impact_to_simplify=open("impact_to_simplify.smt","w")
   impact_to_simplify.write(vars_declaration.read())
   vars_declaration.close()
   impact_to_simplify.write(interv_declaration.read())
   interv_declaration.close()
   impact_to_simplify.write(impact_declaration.read())
   impact_declaration.close()
   impact_to_simplify.close()
   call('z3 impact_to_simplify.smt > result_simplif_impact',  env=path, shell=True)

   ##Read result of simplification
   result_simpl_impact="result_simplif_impact"
   (lines) = utils.read_lines(result_simpl_impact)
   impact_simple_first=lines[2:-2]
   impact_simple_before=copy.copy(impact_simple_first)
#   print("\nFirst simplified formula contains: "+str(len(impact_simple_first))+" lines which are:")
#   print(list(impact_simple_before))
   ## If the predicate to simplify has more parts than variables (at least a or operator)

   if len(impact_simple_first)>1 and ("or" in impact_simple_first[0] or "and" in impact_simple_first[0]):
#   if len(impact_simple_first)>(nb_var+nb_interv):
#      print("\nFirst simplification potentially not minimal")
      impact_simple_after=[]
#      print(list(impact_simple_before))
##      ##while the result of the simplification is different from the previous one
      while True:
#         ##Create new model to simplify from the first simplification result
         open("impact_to_simplify.smt","w").close()
         vars_declaration=open("vars_declaration.smt","r")
         interv_declaration=open("interv_declaration.smt", "r")
         to_simpl=open("impact_to_simplify.smt","w")
         to_simpl.write(vars_declaration.read())
         vars_declaration.close()
         to_simpl.write(interv_declaration.read())
         interv_declaration.close()
         to_simpl.write("(assert \n")
         for l in range(len(impact_simple_before)):
#            print(impact_simple_before[l])
            to_simpl.write(str(impact_simple_before[l]))
         to_simpl.write(")\n(apply ctx-solver-simplify) ")
         to_simpl.close()
         size_before=len(impact_simple_before)
         call('z3 impact_to_simplify.smt > result_simplif_impact', env=path,  shell=True)
         result_simpl_impact="result_simplif_impact"
         (lines) = utils.read_lines(result_simpl_impact)
         impact_simple_after=lines[2:-2]
         size_after=len(impact_simple_after)
         impact_simple_before=copy.copy(impact_simple_after)
         impact_simple_after=[]
         if size_before==size_after:
            break
   impact_simple=copy.copy(impact_simple_before)
#   print(list(impact_simple))
#   print("----------------------End simplification-------------------------")
   open('impact_to_implify.smt',"w").close()
   vars_declaration=open("vars_declaration.smt","r")
   interv_declaration=open('interv_declaration.smt','r')
   to_simpl=open('impact_to_implify.smt',"w")
   to_simpl.write(vars_declaration.read())
   vars_declaration.close()
   impact_declaration=open("impact_declaration.smt", "r")
   to_simpl.write(impact_declaration.read())
   impact_declaration.close()
   if any('and' in r for r in impact_simple) or any('or' in r for r in impact_simple):
      to_simpl.write("(assert \n")
      for l in range(len(impact_simple)):
         #            print(result_simpl_simple_before[l])
         to_simpl.write(str(impact_simple[l]))
      to_simpl.write(')\n')
   else:
      for l in range(len(impact_simple)):
         to_simpl.write('(assert \n\t'+impact_simple[l]+')\n')
   to_simpl.write("(apply ctx-solver-simplify) ")
   to_simpl.close()
   return impact_simple ##A list of the lines of the simplified formula (smt format)

def simplif_impact_parse_result_smt(impact_simple):
   print("----------------------Begin parse result smt-------------------------")   
   exists_or=0
   line_checked=0
   line_and=[]
   formula_left=[]
   formula_right=[]
   formula_lines_structure=[]
   and_on_several_lines=0
   or_on_one_lines=0
   line_atoms_left=[]
   line_atoms_right=[]
#   print(impact_simple)
   if "or" in impact_simple[0]:
#      print("First line starts with or")
      exists_or=1;
      impact_simple[0]=impact_simple[0].split("or")[-1]
   for j in range(len(impact_simple)):
      formula_line=FormulaLine(exists_or, 0, [],[])
#      print("\n Analyzine line: "+impact_simple[j].lstrip())

      if exists_or==1 and "and" in impact_simple[j] and impact_simple[j][0:3]!="and":
#         print("OR on one line")
         or_on_one_lines=1
         formula_line.exists_and=1
#         if impact_simple[j].count("and")>1:
#            print("More than one AND in line, TO IMPLEMENT")
         formula_left=impact_simple[j].split("(and")[0]
         formula_right=impact_simple[j].split("(and")[-1]
#         print(formula_left+"    OR    "+formula_right)
         for f in formula_left:
            line_atoms_left=formula_left.split("(=")[1:]
#         print(line_atoms_left)
         for f in formula_right:
            if formula_right.count("(=")>1:
               line_atoms_right=formula_right.split("(=")[1:]
#         print(line_atoms_right)
         line_checked=1
      if impact_simple[j][0:3]=="and":
#         print("Line starts with and")
         formula_line.exists_and=1
         impact_simple[j]= impact_simple[j].split("and")[-1].lstrip()
         if "))\n" in impact_simple[j]: #and on one line
#            print("The and is on one line")
            impact_simple[j]=impact_simple[j].split("))")[0]
#            print(impact_simple[j])
            line_atoms=impact_simple[j].split("(= ")[1:]
            line_checked=1         
         if impact_simple[j][-2:-1]!="))":
#            print("and on several lines")
            and_on_several_lines=1
      if "and" not in impact_simple[j] and len(impact_simple)>1:
            and_on_several_lines=1
      if and_on_several_lines==1:
#         print("Adding line to the previous line of the AND:")
         line_and.append( impact_simple[j])
#         print(line_and)
         if impact_simple[j][-3:-1]=="))":
            line_atoms=' '.join(line_and)
            line_atoms=line_atoms.split("(= ")[1:]
            line_checked=1
#            print("Total line checked, line_atoms:")
#            print(list(line_atoms))


         
         
      if formula_line.exists_and==0 and formula_line.exists_or==0:
         line_checked=0
#         print("No operator in line")
         if j!=0:
            impact_simple[0]+= impact_simple[j]
#         print("Flag interv simplified: \n"+impact_simple[0])
         if j==len(impact_simple)-1:
            line_atoms=impact_simple[0].split("(= ")[1:]
            line_checked=1         
      if line_checked==1:
         line_and=[]
         if or_on_one_lines==1:
            formula_line.names=[]
            formula_line.values=[]
            for l in  range(len(line_atoms_left)):
               while ")" in line_atoms_left[l]:
                  line_atoms_left[l]=line_atoms_left[l].split(")")[0]
#               print("Atom line left: "+ line_atoms_left[l])
               line_atoms_left_names=line_atoms_left[l].lstrip().split(" ")[0]
#               print("Name in atom: "+line_atoms_left_names)
               line_atoms_left_val=line_atoms_left[l].split(" "+line_atoms_left_names)[-1]
#               print("Value in atom: "+line_atoms_left_val)
               formula_line.names.append(line_atoms_left_names)
               formula_line.values.append(line_atoms_left_val)
            formula_lines_structure.append(formula_line)
#            for f in formula_lines_structure:
#               print("Stored in structure (name): ")
#               print(list(f.names))
#               print("Stored in structure (value): ")
#               print(list(f.values))
            formula_line.names=[]
            formula_line.values=[]
            for l in  range(len(line_atoms_right)):
               while ")" in line_atoms_right[l]:
                  line_atoms_right[l]=line_atoms_right[l].split(")")[0]
#               print(line_atoms_right[l])               
               line_atoms_right_names=line_atoms_right[l].lstrip().split(" ")[0]
#               print(line_atoms_right_names)
               line_atoms_right_val=line_atoms_right[l].split(" "+line_atoms_right_names)[-1]
               formula_line.names.append(line_atoms_right_names)
               formula_line.values.append(line_atoms_right_val)
            formula_lines_structure.append(formula_line)
#            for f in formula_lines_structure:
#               print("Stored in structure (name): ")
#               print(list(f.names))
#               print("Stored in structure (value): ")
#               print(list(f.values))
         if or_on_one_lines!=1:   
            for l in range(len(line_atoms)):
               line_atoms[l]=line_atoms[l].split(")")[0]
               line_atom_name_l=line_atoms[l].split(" ")[0]
               line_atom_smt_val_l=line_atoms[l].split(" ")[1]
               line_atom_val_l=line_atom_smt_val_l.split(line_atom_name_l)[-1]
#               print("In "+str(l)+" atom of line, the name is "+line_atom_name_l+" and the value is "+line_atom_val_l)
               formula_line.names.append(line_atom_name_l)
               formula_line.values.append(line_atom_val_l)
#               print("Stored in the structure: ")
#               print( list(formula_line.names))
#               print(list(formula_line.values))
            line_checked=0
            formula_lines_structure.append(formula_line) 
#         for f in formula_lines_structure:
#            print("Stored formula: ")
#            print(str(f.exists_or)+"\t"+str(f.exists_and))
#            print(list(f.names))
#            print(list(f.values))
   print("----------------------End parse result smt-------------------------")            
   return formula_lines_structure

def simplif_strategy_write_result_smv(formula_lines_structure, fct_name , model_name, names_intervs):
   result_impact_smv=[]
   exists_or=0
   for f in formula_lines_structure:
      if f.exists_or==1:
         exists_or=1
      line_atom=""
#      print(list(f.names))
#      print(list(f.values))
      for n in range(len(f.names)):
         if n<len(f.names)-1:
            if any(f.names[n] in n_i for n_i in names_intervs):
               line_atom+=f.names[n]+".activ = "+f.values[n]+" & "
            else:
               line_atom+=f.names[n]+".v = "+f.values[n]+" & "
         if n==len(f.names)-1:
            if any(f.names[n] in n_i for n_i in names_intervs):
               line_atom+=f.names[n]+".activ = "+f.values[n]
            else:
               line_atom+=f.names[n]+".v = "+f.values[n]
      result_impact_smv.append(line_atom)
#      print("Stored result: ")
#      print(list(result_impact_smv))
   result_smv=open(model_name+".impact.smv","a")
   result_smv.write("\nDEFINE non_reach_"+fct_name+" :=")
   for j in range(len(result_impact_smv)):
      result_smv.write(" ("+result_impact_smv[j]+") ")
      if exists_or==1:
         if j<len(result_impact_smv)-1:
            result_smv.write(" | ")
      if j==len(result_impact_smv)-1:
         result_smv.write(";\n")
   result_smv.close()